#include "sprfile.h"

/*
    SPRFile.SoftBank
    ________________________________________________________________________

    The SoftBank class is intended for use with .spr files from
    SoftBank's port of the Neo*Geo game, Shin Samurai Spirits (aka SS2).
    This class facilitates the extraction and viewing of the sprites
    contained within that game.

    Please have respect for the art you extract.

    Copyright (C) 1997 / John Beppu
    Released under the GNU Public License
 
    ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::' */

struct SPRString {
    u32         len;
    u8          *data;
};

struct SPRite {
    u32         wd;
    u32         hi;
    u32         offset;
    SPRString   *strLUT;
};

class SoftBank : public SPRFile {

public:

    SoftBank (
        i8 *filename
        );
   ~SoftBank();

    u32 spriteWd (
        u32 spriteNDX
        ) const;
    u32 spriteHi (
        u32 spriteNDX
        ) const;
    u32 sizeOfHSI (
        u32 spriteNDX
        ) const;
    void makeHSI (
        u8 *dst,
        u32 spriteNDX
        ) const;
    void blt8bpp (
        u8 *dst,
        u32 pitch,
        u32 spriteNDX
        ) const;

private:

    u32         fileLEN;
    FILE        *file0;

    u32         kakaFactor;
    u32         bufferLEN;
    u8          *buffer;
    //          spriteQTY;      // protected SPRFile member
    SPRite      *spriteLUT;

    void checkSignature();
    u32 skipBullShit();
    void findSpriteQty(u32 ndx);
    void loadSpriteHeaders(u32 ndx);
    void loadSprites (
        i8 *filename,
        u32 ndx
        );

};

// Methods :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

// #define SOFTBANK_DEBUG
#define BAD_FORMAT      1

/*
    [function]  Checks whether the current buffer contains the
                2 byte signature common to all SoftBank files.
`.....................................................................*/
void SoftBank::checkSignature()
{
    u32 u0 = 0;
    u0 |= (u32) (buffer[0] | buffer[1] << 8);
    if (u0 != 0x00001053) throw(BAD_FORMAT);
}


/*
    [function]  This skips past all the seemingly unecessary data
                at the beginning of many .spr files.
    [returns]   The index into buffer where meaningful data begins.
`.....................................................................*/
u32 SoftBank::skipBullShit()
{
    u32 check0,check1;
    u32 u0,u1;
    u32 *u32_buffer = (u32*) buffer;
    u32 ndx;
    u32 limit = (bufferLEN >> 2);

    #ifdef BIG_ENDIAN
    check0 = 0x01000000;
    #else
    check0 = 0x00000001;
    #endif
    check1 = 0;

    // #define SKIP_MAGIC          // comment this line for slow way
    #ifdef SKIP_MAGIC

    u32     skipFactor;
    u32     skipNDX;
    float   skipMagic = 11.9;

    skipFactor = buffer[2] | (buffer[3] << 8);
    skipNDX = u32 (skipFactor * skipMagic);
    while ((skipNDX - 8) % 12) skipNDX++;
    ndx = skipNDX >> 2;

    #else

    ndx = 2;

    #endif

    do {
        u0 = u32_buffer[ndx];
        u1 = u32_buffer[ndx+1];
        u0 -= check0;
        u1 |= u32_buffer[ndx+2];
        ndx += 3;
        u0 |= u1;
    } while ((u0 == 0) && (ndx < limit));

    if (ndx >= limit) return 0;

    ndx -= 3;
    ndx = ndx << 2;
    return ndx;
}


/*
    [pre]       Skip bullshit first.
    [function]  Finds how many sprites are in the file
`.....................................................................*/
void SoftBank::findSpriteQty(u32 ndx)
{
    u32 check0,check1;
    u32 signature;
    u32 *u32_buffer;
    u32 index;
    u32 u0,u1;

    #ifdef BIG_ENDIAN
    check0 = 0x01000000;
    check1 = 0xffff0000;
    #else
    check0 = 0x00000001;
    check1 = 0x0000ffff;
    #endif

    u32_buffer = (u32*) buffer;
    index = ndx >> 2;
    u1 = 0;

    signature = u32_buffer[index];
    u0 = u32_buffer[index];
    while ((u0 & check1) == 0) {
        index += 3;
        u1++;
        u0 = u32_buffer[index];
        if (u0 == check0) {
            this->kakaFactor++;
            u0 = signature;
            u1--;
        }
    }
    spriteQTY = u1;
}


/*
    [function]
`.....................................................................*/
void SoftBank::loadSpriteHeaders(u32 ndx)
{
    u32 u0,u1,u2;
    u32 *u32_buffer;
    u32 index;
    u32 check0;

    #ifdef BIG_ENDIAN
    check0 = 0x01000000;
    #else
    check0 = 0x00000001;
    #endif

    u32_buffer = (u32*) buffer;

    // mallocate some memory :::::::::::::::::::::::::::::::::::::::::::

    spriteLUT = (SPRite*) malloc(spriteQTY * sizeof(SPRite));
    if (spriteLUT == NULL) return;

    // load headers and allocate strLUT's ::::::::::::::::::::::::::::::

    for (u0 = 0; u0 < spriteQTY; u0++) {

        spriteLUT[u0].wd = buffer[ndx+4] | (buffer[ndx+5] << 8);
        spriteLUT[u0].hi = buffer[ndx+6] | (buffer[ndx+7] << 8);
        spriteLUT[u0].offset = buffer[ndx+8];
        spriteLUT[u0].offset |= buffer[ndx+9] << 8;
        spriteLUT[u0].offset |= buffer[ndx+10] << 16;
        spriteLUT[u0].offset |= buffer[ndx+11] << 24;
        ndx += 12;

        index = ndx >> 2;
        u2 = u32_buffer[index];
        while (u2 == check0) {
            index += 3;
            u2 = u32_buffer[index];
        }
        ndx = index << 2;

        #ifdef SOFTBANK_DEBUG

        printf("sprite %d\n",u0);
        printf("    wd = %08x   hi = %08x\n",
            spriteLUT[u0].wd,
            spriteLUT[u0].hi);
        printf("    offset = %08x\n",spriteLUT[u0].offset);
        #endif

        u1 = spriteLUT[u0].hi;
        spriteLUT[u0].strLUT =
            (SPRString*) malloc(u1 * sizeof(SPRString));
        if (spriteLUT[u0].strLUT == NULL) {
	    return;
	}
    }
}


/*
    [function]
`.....................................................................*/
void SoftBank::loadSprites (
    i8 *filename,
    u32 ndx
    )
{
    u32 shinBufferLEN;
    u32 scanlineQTY;
    u32 u0,u1,u2,u3;

    SPRite  *spr0;

    // trying not to waste memory ::::::::::::::::::::::::::::::::::::::

    shinBufferLEN = fileLEN - ndx;
    buffer = (u8*) realloc(buffer,shinBufferLEN);
    file0 = fopen(filename,"rb");
    fseek(file0,ndx,SEEK_SET);
    fread(buffer,sizeof(u8),shinBufferLEN,file0);
    bufferLEN = shinBufferLEN;
    fclose(file0);

    // the final stage of sprite loading :::::::::::::::::::::::::::::::

    for (u0 = 0; u0 < spriteQTY; u0++) {
        spr0 = &spriteLUT[u0];
        u2 = spr0->offset;
        scanlineQTY = spr0->hi;
        #ifdef SOFTBANK_DEBUG
        printf("sprite %d\n",u0);
        printf("  offset : %08x\n",u2);
        printf("  height : %08x\n",scanlineQTY);
        printf("  width  : %08x\n",spr0->wd);
        #endif
        for (u1 = 0; u1 < scanlineQTY; u1++) {
            u3 = buffer[u2] | (buffer[u2+1] << 8);
            spr0->strLUT[u1].data = &buffer[u2+2];
            u2 += u3;
            spr0->strLUT[u1].len = u3;
            #ifdef SOFTBANK_DEBUG
            {
                printf("  lineLen[%3d] = %3d",u1,u3);
                if (((u1+1) % 4) == 0) printf("\n");
            }
            #endif
        }
        #ifdef SOFTBANK_DEBUG
        printf("\n");
        getchar();
        #endif
    }
}


/*
    [function]  Loads the specified spr file.
`.....................................................................*/
SoftBank::SoftBank (
    i8 *filename
    )
    :SPRFile()
{
    kakaFactor = 0;
    file0 = fopen(filename,"rb");
    if (file0 == NULL) return;
    fseek(file0,0,SEEK_END);
    fileLEN = ftell(file0);
    fseek(file0,0,SEEK_SET);
    bufferLEN = fileLEN;
    buffer = (u8*) malloc(bufferLEN * sizeof(u8));
    fread(buffer,sizeof(u8),bufferLEN,file0);
    fclose(file0);

    u32 bNDX;
    u32 u0;

    checkSignature();
    bNDX = skipBullShit();
    findSpriteQty(bNDX);
    loadSpriteHeaders(bNDX);
    u0 = bNDX + ((spriteQTY + kakaFactor) * 12);
    #ifdef SOFTBANK_DEBUG
    printf("\n[0x%08x] The perceived base address\n",u0);
    getchar();
    #endif
    loadSprites(filename,u0);
}


/*
    [function]  Deallocates memory
`.....................................................................*/
SoftBank::~SoftBank()
{
    u32         u0;
    SPRString   *str0;

    #if 1
    if (spriteLUT != NULL) {
        for (u0 = 0; u0 < spriteQTY; u0++) {
            str0 = spriteLUT[u0].strLUT;
            if (str0 != NULL) free(str0);
        }
        free(spriteLUT);
    }
    #endif
    if (buffer != NULL) free(buffer);
}


/*
    [returns]   Width of sprite.
`.....................................................................*/
u32 SoftBank::spriteWd (
    u32 spriteNDX
    ) const
{
    if (spriteNDX > spriteQTY) throw("INVALID_SPRITE");
    return spriteLUT[spriteNDX].wd;
}


/*
    [returns]   Height of sprite.
`.....................................................................*/
u32 SoftBank::spriteHi (
    u32 spriteNDX
    ) const
{
    if (spriteNDX > spriteQTY) throw("INVALID_SPRITE");
    return spriteLUT[spriteNDX].hi;
}



/*
    [returns]   The amount of memory (in bytes) required to store
                an HSI-raw version of the sprite.
`.....................................................................*/
u32 SoftBank::sizeOfHSI (
    u32 spriteNDX
    ) const
{
    u32 size;
    if (spriteNDX >= spriteQTY) return 0;
    size = HSI_HEADER_LEN;
    size += palLEN * 3;
    size += (spriteLUT[spriteNDX].wd * spriteLUT[spriteNDX].hi);
    size *= sizeof(u8);
    return size;
}


/*
    [function]  Make an HSI-raw version of the specified sprite
                at dst[]
`.....................................................................*/
void SoftBank::makeHSI (
    u8 *dst,
    u32 spriteNDX
    ) const
{
    static u8 HSI_Header[32] = {
        0x6d, 0x68, 0x77, 0x61, 0x6e, 0x68,     // magic number
        0x00, 0x04,                             // version
        0x00, 0x00,                             // width
        0x00, 0x00,                             // height
        0x00, 0x00,                             // palette size
        0x00, 0x00,                             // horizontal dpi
        0x00, 0x00,                             // vertical dpi
        0x00, 0x00,                             // gamma
        0x00, 0x00,                             // compression
        0x00, 0x00, 0x00, 0x00, 0x00,           // reserved
        0x00, 0x00, 0x00, 0x00, 0x00            // reserved
    };

    SPRite  *spr0;
    u32     u0;

    u0 = spriteNDX;
    if (spriteNDX >= spriteQTY) return;
    spr0 = &spriteLUT[spriteNDX];

    // copy header :::::::::::::::::::::::::::::::::::::::::::::::::::::

    if (palLEN == 0) {
        // genPalette()
    }
    memcpy(dst,HSI_Header,HSI_HEADER_LEN * sizeof(u8));

    // big-endian style.  (grief for little-endians)

    dst[8]  = (u8) (u32 (spr0->wd & 0x0000ff00) >> 8);
    dst[9]  = (u8) (u32 (spr0->wd & 0x000000ff));
    dst[10] = (u8) (u32 (spr0->hi & 0x0000ff00) >> 8);
    dst[11] = (u8) (u32 (spr0->hi & 0x000000ff));
    dst[12] = (u8) (u32 ((palLEN & 0x0000ff00) >> 8));
    dst[13] = (u8) (palLEN & 0x000000ff);
    dst += HSI_HEADER_LEN * sizeof(u8);

    // copy palette ::::::::::::::::::::::::::::::::::::::::::::::::::::

    memcpy(dst,palBUF,(palLEN * 3) * sizeof(u8));
    dst += (palLEN * 3) * sizeof(u8);

    // copy sprite over pal[0] background ::::::::::::::::::::::::::::::

    u0 = spr0->wd * spr0->hi * sizeof(u8);
    memset(dst,0,u0);
    blt8bpp(dst,spr0->wd,spriteNDX);
}


/*
    [function]  Block-transfer the specified sprite to dst[].
`.....................................................................*/
void SoftBank::blt8bpp (
    u8 *dst,
    u32 pitch,
    u32 spriteNDX
    ) const
{
    u8      *scanline;
    u32     height;
    u32     u0,u1,u2,u3;
    SPRite  *spr0;

    u8      *instPTR;           // instruction pointer
    u32     h_OFS;              // horizontal offset
    u32     strLEN;             // string length
    u8      *dataPTR;           // data pointer

    if (spriteNDX >= spriteQTY) return;

    spr0 = &spriteLUT[spriteNDX];
    scanline = dst;
    height = spr0->hi;
    for (u0 = 0; u0 < height; u0++) {
        dst += pitch;
        instPTR = spr0->strLUT[u0].data;

        if (spr0->strLUT[u0].len > 4) {         // tricky.
            u2 = 2;
            while (spr0->strLUT[u0].len > u2) {
                u1 = (u32) instPTR[0];
		u3 = 0;
                if (u1 > 0x7f) {
		    instPTR--; 
		}
                h_OFS = (u32) instPTR[0];
                strLEN = (u32) (instPTR[1] - 0x80);
                dataPTR = &instPTR[2];

		if ((i32) strLEN <= 0) {
                    #if 0
		    i32 i0,i1 = 1;
		    u8  *dump = &instPTR[0] - 4;
		    printf("\n");
		    for (i0 = 80; i0 != 0; i0--) {
			printf("%02x ",(*dump++));
			if (i1 % 16 == 0) printf("\n");
			i1++;
		    }
		    printf("\n");
                    #endif
		    u2 = 0xffffffff;
		}
		else {
                    scanline += h_OFS;
                    instPTR += (strLEN + 3);
                    u2 += (strLEN + 3);
		    u1 = strLEN;
		    do {
                        (*scanline++) = (*dataPTR++);
			u1--;
		    } while ((i32) u1 > 0);
		}
            }
        }
        scanline = dst;
    }
}

